/*
 * The Kibitz - Simple HTTP daemon implementation
 *
 * (c) 2005-2012 Tomas Valla <tom@ucw.cz>
 *
 * This software may be freely distributed and used according to the terms
 * of the GNU Lesser General Public License. See COPYING for details.
 */

#undef LOCAL_DEBUG

#include <kibitz/lib.h>
#include <ucw/clists.h>
#include <ucw/conf.h>
#include <ucw/mainloop.h>
#include <ucw/mempool.h>
#include <ucw/stkstring.h>
#include <ucw/url.h>
#include <kibitz/internal.h>
#include <charset/charconv.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <time.h>
#include <unistd.h>

#define CLOG(c,l,t,m,a...) do { if (config.log_connections >= (l))     \
	   msg(t, "Connection %d (%s): " m, (c)->rio.file.fd, (c)->ip_addr, a); } while(0)

static struct main_file listening_socket;

static struct clist conn_list;

struct query_word {
	cnode n;
	char *word;
	double bonus;
};

static void do_reply(struct connection *c, int code, char *reason)
{
	if (c->replied)
		return;
	time_t t = time(NULL);
	struct tm *tm = gmtime(&t);
	char tm_buf[100];
	strftime(tm_buf, sizeof(tm_buf)-1, "%a, %d %b %Y %H:%M:%S %Z", tm);

	c->ignore_input = 1;
	c->replied = 1;
	c->reply_buf = mp_printf(c->mp, "HTTP/1.%d %d %s\r\nDate: %s\r\n%s\r\n", c->http_subver, code, reason, tm_buf,
		(c->http_subver == 1 ? "Connection: close\r\n" : "") );

	rec_io_set_timeout(&c->rio, config.response_timeout);
	rec_io_write(&c->rio, c->reply_buf, strlen(c->reply_buf));
	CLOG(c, 2, L_INFO, "Replying status message %d %s", code, reason);
}

static void reply_bad_request(struct connection *c)
{
	DBG("Reply bad request conn=%d", c->rio.file.fd);
	do_reply(c, 400, "Bad Request");
}

static void do_query(struct connection *c)
{
	DBG("Executing query for conn=%d", c->rio.file.fd);

	c->q = kibitz_new_query();
	kibitz_init_query(c->q);

	struct query_word *qw;
	CLIST_WALK(qw, c->query_words) {
		kibitz_add_expr(c->q, qw->word, strlen(qw->word), qw->bonus);
	}
	if (c->result_cnt && (uns)c->result_cnt < config.query_max_results)
		c->q->max_results = c->result_cnt;
	else
		c->q->max_results = config.query_max_results;

	kibitz_get_results(c->q);
#ifdef LOCAL_DEBUG
	kibitz_dump_query(c->q);
#endif
}

static char *charset_conv(struct mempool *mp, int ie, int oe, char *src, uns srclen)
{
	struct conv_context cc;
	conv_init(&cc);
	conv_set_charset(&cc, ie, oe);

	uns nl = srclen;
	char *nr;
	int flg;
	do {
		nl *= 2;
		nr = mp_alloc(mp, nl);
		cc.source = src;
		cc.source_end = src + srclen;
		cc.dest = cc.dest_start = nr;
		cc.dest_end = nr + nl - 1;
		flg = conv_run(&cc);
	} while ((flg & CONV_DEST_END) && !(flg & CONV_SOURCE_END));
	*cc.dest = 0;

	return nr;
}

static void reply_results(struct connection *c)
{
	DBG("Reply results conn=%d", c->rio.file.fd);

	char *rbuf[c->q->num_results + 2];
	rbuf[0] = CUSTOM_HTTP_RESULT_HEADER(c);
	uns i = 1;

	if (c->q->num_results) {
		struct kibitz_result *r;
		CLIST_WALK(r, c->q->results) {
			rbuf[i] = CUSTOM_HTTP_RESULT_LINE(c, r);
			if (i != c->q->num_results)
				rbuf[i] = stk_strcat(rbuf[i], CUSTOM_HTTP_RESULT_LINE_DELIM(c));
			i++;
		}
	}
	rbuf[i++] = CUSTOM_HTTP_RESULT_FOOTER(c);
	char *results = stk_strarraycat(rbuf, i);
	uns reslen = strlen(results);

	int oe;
	if (c->output_enc && (oe = find_charset_by_name(c->output_enc)) >= 0) {
		if (oe != CONV_CHARSET_UTF8) {
			results = charset_conv(c->mp, CONV_CHARSET_UTF8, oe, results, reslen);
			reslen = strlen(results);
		}
	}
	else {
		CLOG(c, 2, L_WARN, "Output encoding '%s' missing or invalid, defaulting to UTF8", c->output_enc);
		c->output_enc = mp_strdup(c->mp, "utf-8");
	}

	time_t t = time(NULL);
	struct tm *tm = gmtime(&t);
	char tm_buf[100];
	strftime(tm_buf, sizeof(tm_buf)-1, "%a, %d %b %Y %H:%M:%S %Z", tm);

	c->reply_buf = mp_printf(c->mp, "HTTP/1.%d 200 OK\r\nDate: %s\r\n%sContent-Type: text/plain; charset=%s\r\n"
		"Content-Length: %d\r\n\r\n", c->http_subver, tm_buf, (c->http_subver == 1 ? "Connection: close\r\n" : ""),
		c->output_enc, reslen);

	if (c->get_method)
		c->reply_buf = mp_strcat(c->mp, c->reply_buf, results);

	rec_io_set_timeout(&c->rio, config.response_timeout);
	rec_io_write(&c->rio, c->reply_buf, strlen(c->reply_buf));
	CLOG(c, 2, L_INFO, "Replying %d results (%d bytes total)", c->q->num_results, reslen);
}

/* 1 on success, 0 on error */
static int load_argbuf(struct connection *c, char *ubuf)
{
	char ubuf1[MAX_URL_SIZE], ubuf2[MAX_URL_SIZE];

	/* Substitute + for space */
	char *s = ubuf;
	while (*s) {
		if (*s == '+')
			*s = ' ';
		s++;
	}

	int r = url_deescape(ubuf, ubuf1);
	if (r) {
		CLOG(c, 1, L_WARN, "Error deescaping URL '%s': %s", ubuf, url_error(r));
		if (r == URL_ERR_TOO_LONG)
			do_reply(c, 414, "Request-URI Too Long");
		else
			reply_bad_request(c);
		return 0;
	}
	struct url url;
	bzero(&url, sizeof(url));
	r = url_split(ubuf1, &url, ubuf2);
	if (r) {
		CLOG(c, 1, L_WARN, "Error splitting URL '%s': %s", ubuf1, url_error(r));
		if (r == URL_ERR_TOO_LONG)
			do_reply(c, 414, "Request-URI Too Long");
		else
			reply_bad_request(c);
		return 0;
	}
	char *p = url.rest;
	if (!*p || *p != '/' || *(p+1) != '?') {
		CLOG(c, 1, L_WARN, "Malformed URL '%s'", ubuf);
		do_reply(c, 404, "Not Found");
		return 0;
	}
	p += 2;

	c->arg_buf = mp_strdup(c->mp, p);
	return 1;
}

/* Return 0 on empty line, 1 otherwise */
static int parse_request(struct connection *c, byte *s)
{
	DBG("Parsing request, conn=%d '%s'", c->rio.file.fd, s);
	byte *os = s;

	/* Take care of the empty line */
	if (!*s || *s == '\r' || *s == '\n')
		return 0;

	if (c->ignore_input)
		return 1;

	int get = strncmp(s, "GET ", 4) == 0;
	int head = strncmp(s, "HEAD ", 5) == 0;

	static byte *mlist[] = { "OPTIONS ", "POST ", "DELETE ", "TRACE ", "CONNECT " };
	int other = 0;
	for (uns i=0; i<5; i++)
		other += strncmp(s, mlist[i], strlen(mlist[i])) == 0;

	if (other) {
		CLOG(c, 1, L_WARN, "Request '%s': method not implemented", os);
		do_reply(c, 501, "Not Implemented");
		return 1;
	}
	if (!(get+head)) {
		CLOG(c, 1, L_WARN, "Request '%s': bad method", os);
		reply_bad_request(c);
		return 1;
	}

	/* Separate the URL */
	while (*s && *s != ' ')
		s++;
	s++;
	char *us = s;
	while (*s && *s != ' ')
		s++;
	char *ue = s;
	if (!load_argbuf(c, stk_strndup(us, ue-us)))
		return 1;

	s++;
	if (strncmp(s, "HTTP/", 5) != 0) {
		CLOG(c, 1, L_WARN, "Request '%s': bad request", os);
		reply_bad_request(c);
		return 1;
	}
	s += 5;
	if (strncmp(s, "1.0", 3) == 0)
		c->http_subver = 0;
	else if (strncmp(s, "1.1", 3) == 0)
		c->http_subver = 1;
	else {
		c->http_subver = 1;
		CLOG(c, 1, L_WARN, "Request '%s': unsupported HTTP version", os);
		do_reply(c, 505, "Version not supported");
		return 1;
	}
	s += 3;
	if (*s && *s != '\n' && *s != '\r') {
		CLOG(c, 1, L_WARN, "Request '%s': bad request", os);
		reply_bad_request(c);
		return 1;
	}
	if (get)
		c->get_method = 1;
	c->ignore_input = 1;
	return 1;
}

static void insert_word(struct connection *c, char *w, double b)
{
	struct query_word *qw = mp_alloc_zero(c->mp, sizeof(*qw));
	qw->word = mp_strdup(c->mp, w);
	qw->bonus = b;
	clist_add_tail(&c->query_words, &qw->n);
}

/* 1 on success, 0 otherwise */
static int load_query_words(struct connection *c)
{
	if (!c->qw_buf)
		return 0;
	char *w = stk_strdup(c->qw_buf);
	DBG("Loading query words '%s'", w);

	clist_init(&c->query_words);

	/* Handle empty word */
	if (!*w) {
		insert_word(c, w, config.single_word_bonus);
		return 1;
	}

	/* Normalize the spaces and escaped chars */
	char *s = w;
	char *d = w;
	while (*s && *s == ' ')
		s++;
	if (!*s)
		return 0;
	while (*s) {
		if ((uns)*s < NCC_MAX)
			*s = NCC_CHARS[(uns)*s];

		if (*s != ' ' || *(s+1) != ' ')
			*d++ = *s;
		s++;
	}
	*d = 0;

	int ie;
	if (c->input_enc && (ie = find_charset_by_name(c->input_enc)) >= 0) {
		if (ie != CONV_CHARSET_UTF8)
			w = charset_conv(c->mp, ie, CONV_CHARSET_UTF8, w, strlen(w));
	}
	else
		CLOG(c, 2, L_WARN, "Input encoding '%s' missing or invalid, defaulting to UTF8", c->input_enc);

	int cnt = 0;
	char *ow = stk_strdup(w);
	d = s = w;
	while (*s) {
		if (*s == ' ') {
			*s = 0;
			insert_word(c, d, config.single_word_bonus);
			cnt++;
			d = s+1;
		}
		s++;
	}
	if (d != w && *d)
		insert_word(c, d, config.single_word_bonus);

	if (cnt)
		insert_word(c, ow, config.phrase_bonus);
	else
		insert_word(c, ow, config.single_word_bonus);


	return 1;
}

/* Return 1 on success, 0 otherwise */
static int parse_argbuf(struct connection *c)
{
	DBG("Parsing argbuf '%s' conn=%d", c->arg_buf, c->rio.file.fd);

	if (!c->arg_buf)
		return 0;

	char *b, *v, *s = c->arg_buf;
	int endloop = 0;

	for (;;) {
		/* Get the variable name */
		b = s;
		while (*s && *s != '=')
			s++;
		if (!*s)
			return 0;
		*s = 0;

		/* Get the value */
		v = s+1;
		while (*v && *v != '&')
			v++;
		if (!*v)
			endloop = 1;
		*v = 0;

		s++;
		DBG("b='%s' s='%s'", b, s);
		if (!strcmp(b, "idx"))
			c->index_name = mp_strdup(c->mp, s);
		else if (!strcmp(b, "q"))
			c->qw_buf = mp_strdup(c->mp, s);
		else if (!strcmp(b, "ie"))
			c->input_enc = mp_strdup(c->mp, s);
		else if (!strcmp(b, "oe"))
			c->output_enc = mp_strdup(c->mp, s);
		else if (!strcmp(b, "of"))
			c->output_format = atoi(s);
		else if (!strcmp(b, "debug"))
			c->debug_mode = 1;
		else if (!strcmp(b, "cnt")) {
			char *e;
			c->result_cnt = strtol(s, &e, 10);
			if (!*s || *e)
				return 0;
		}
		else
			return 0;
		s = v+1;

		if (endloop)
			break;
	}
	CLOG(c, 1, L_INFO, "Request q='%s' idx='%s' ie='%s' oe='%s' debug=%d cnt=%d", c->qw_buf, c->index_name, c->input_enc, c->output_enc, c->debug_mode, c->result_cnt);
	return 1;
}

static void parse_line(struct connection *c, byte *st, byte *e)
{
	*e = 0;
	DBG("Connection %d, parsing line '%s' (%ld bytes)", c->rio.file.fd, st, e-st);

	char *s = stk_strdup(st);

	if (!parse_request(c, s)) {
		if (!parse_argbuf(c)) {
			CLOG(c, 1, L_WARN, "Error parsing arguments '%s'", c->arg_buf);
			reply_bad_request(c);
			return;
		}
		if (!load_query_words(c)) {
			CLOG(c, 1, L_WARN, "Error parsing query words '%s'", c->qw_buf);
			reply_bad_request(c);
			return;
		}

		do_query(c);
		reply_results(c);
	}
}

static uns conn_read_handler(struct main_rec_io *rio)
{
	uns len = rec_io_parse_line(rio);
	if (!len)
		return 0;

	parse_line(rio->data, rio->read_rec_start, rio->read_rec_start + len - 1);
	return len;
}

static const char *status_str(int status)
{
	switch (status) {
		case RIO_ERR_READ:
			return "read() returned an error";
		case RIO_ERR_WRITE:
			return "write() returned an error";
		case RIO_ERR_TIMEOUT:
			return "A timeout has occurred";
		case RIO_ERR_RECORD_TOO_LARGE:
			return "Read: read_rec_max has been exceeded";
		case RIO_EVENT_ALL_WRITTEN:
			return "All buffered data has been written";
		case RIO_EVENT_PART_WRITTEN:
			return "Some buffered data has been written, but more remains";
		case RIO_EVENT_EOF:
			return "Read: EOF seen";
		default:
			return "Unknown status";
		}
}

static void conn_cleanup(struct connection *c)
{
	rec_io_del(&c->rio);
	close(c->rio.file.fd);
	clist_remove(&c->node);
	if (c->q)
		kibitz_free_query(c->q);
	mp_delete(c->mp);
}

static int conn_notify_handler(struct main_rec_io *rio, int status)
{
	DBG("conn_notify_handler()");
	struct connection *c = rio->data;
	switch (status) {
		case RIO_EVENT_PART_WRITTEN:
		case RIO_EVENT_EOF:
			return HOOK_RETRY;

		case RIO_EVENT_ALL_WRITTEN:
			CLOG(c, 2, L_INFO, "%s", "Closed");
			break;

		case RIO_ERR_TIMEOUT:
			if (c->timer_active) {
				c->timer_active = 0;
				CLOG(c, 1, L_WARN, "%s", "Request timeout");
				do_reply(c, 408, "Request Time-out");
				return HOOK_RETRY;
			}
			/* fallthru */
		default:
			CLOG(c, 1, L_WARN, "Error occurred: %s", status_str(status));
	}

	conn_cleanup(c);
	return HOOK_IDLE;
}

static int handle_new_conn(struct main_file *fi)
{
	DBG("Got new connection");
	struct sockaddr_in sa;
	int len = sizeof(sa);
	int fd = accept(fi->fd, (struct sockaddr *) &sa, &len);
	if (fd < 0) {
		if (errno != EAGAIN && errno != EWOULDBLOCK)
			msg(L_ERROR, "accept: %m");
		return 0;
	}
	if (config.log_connections >= 2)
		msg(L_INFO, "Accepted connection %d from %s", fd, inet_ntoa(sa.sin_addr));

	struct mempool *mp = mp_new(16*1024);
	struct connection *c = mp_alloc_zero(mp, sizeof(*c));
	c->mp = mp;
	c->ip_addr = mp_strdup(mp, inet_ntoa(sa.sin_addr));
	struct main_rec_io *rio = &c->rio;
	rec_io_add(rio, fd);
	clist_add_tail(&conn_list, &c->node);
	rio->read_handler = conn_read_handler;
	rio->notify_handler = conn_notify_handler;
	rec_io_set_timeout(rio, config.request_timeout);
	c->timer_active = 1;
	rio->data = c;
	rec_io_start_read(rio);

	return 1;
}

static void net_init(uns port)
{
	bzero(&listening_socket, sizeof(listening_socket));

	DBG("Initializing listening socket on port %d...", port);
	struct sockaddr_in sa;
	int sk = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	int o = 1;

	if (sk < 0)
		die("socket: %m");
	if (setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &o, sizeof(o)) < 0)
		die("setsockopt(SO_REUSEADDR) failed: %m");

	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(sk, (struct sockaddr *) &sa, sizeof(sa)) < 0)
		die("bind: %m");
	if (listen(sk, config.listen_queue) < 0)
		die("listen: %m");

	listening_socket.fd = sk;
	listening_socket.read_handler = handle_new_conn;
	file_add(&listening_socket);
}

static char index_name[256];
static char config_name[256];

void kibitz_httpd_init(byte *idx_name, byte *cfg_name)
{
	ASSERT(strlen(idx_name) < sizeof(index_name));
	strncpy(index_name, idx_name, sizeof(index_name)-1);
	if (!cfg_name)
		*config_name = 0;
	strncpy(config_name, cfg_name, sizeof(config_name)-1);
	index_name[sizeof(index_name)-1] = config_name[sizeof(config_name)-1] = 0;

	clist_init(&conn_list);

	main_init();

	kibitz_open_index(index_name);

	net_init(config.httpd_port);
	msg(L_INFO, "Daemon initialized, loaded index '%s' and listening on port %d", idx_name, config.httpd_port);

	if (config.httpd_status_file) {
		int sfd = open(config.httpd_status_file, O_WRONLY | O_CREAT | O_TRUNC, 0666);
		if (sfd < 0)
			die("Cannot open status file: %m");
		if (write(sfd, "OK\n", 3) < 0)
			die("Cannot write to status file: %m");
		close(sfd);
	}
}

static void sig_term(struct main_signal *s UNUSED)
{
	msg(L_INFO, "Caught deadly signal, shutting down");
	struct connection *c;
	while (c = clist_remove_head(&conn_list)) {
		CLOG(c, 2, L_INFO, "%s", "Closed by shutdown");
		conn_cleanup(c);
	}
	main_shut_down();
}

static void sig_hup(struct main_signal *s UNUSED)
{
	msg(L_INFO, "Caught SIGHUP, reloading index");
	kibitz_open_index(index_name);
}

static struct main_signal sigs[] = {
	{ .signum = SIGTERM, .handler = sig_term },
	{ .signum = SIGINT, .handler = sig_term },
	{ .signum = SIGHUP, .handler = sig_hup },
};

void kibitz_httpd_exec(void)
{
	DBG("Executing mainloop");
	signal(SIGPIPE, SIG_IGN);
	for (uns i = 0; i < ARRAY_SIZE(sigs); i++)
		signal_add(sigs + i);

	main_loop();

	for (uns i = 0; i < ARRAY_SIZE(sigs); i++)
		signal_del(sigs + i);

	file_del(&listening_socket);
	main_cleanup();
}
