/*
 *	The Kibitz Index File Format
 *
 *	(c) 2005-2012 Tomas Valla <tom@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU Lesser General Public License. See COPYING for details.
 *
 */

#ifndef _KIBITZ_INDEX_H
#define _KIBITZ_INDEX_H

#include <kibitz/lib.h>
#include <custom/custom.h>

/*
 *  The index files have the following format:
 *
 *	header			struct index_header
 *				during the index construction the header is temporary
 *	phrase buffer		struct phrase [num_phrases], aligned structs
 *				the structs have variable lengths
 *	wordoff buffer		u32 [num_words]
 *				for each word the offset of the appropriate phrase
 *				in the phrase buffer
 *	fastfind buffer		struct { u32 poff, u32 quality } [num_fastfind]
 *				precomputed phrase offsets for some trie vertices
 *	trie			the kibitz trie
 */

/*
 * Trie format:
 *
 * The trie disk format is the sequence of compressed trie vertices.
 * String are stored and searched in UTF8, one byte as character, no matter
 * what its real meaning is.
 * Each vertex consists of several elements, some of them optional.
 * The elements are prefixed by an u8 opcode. Elements of one vertex should
 * occur in a fixed order defined lower. In most cases, only one occurence of a
 * given element is allowed.
 * Vertices are referenced via u32 byteoffsets from the start of trie
 * or via u16 relative byteoffsets from the beginning of actual vertex.
 *
 * The opcodes are:
 *
 *	10cccccc	fastfind data
 *			u32 index into fastfind_list; here use cccccc elements
 *
 *	0001cccc	skip character [may occur multiple times, but consecutive]
 *			u8 character [cccc]
 *
 *	01cccccc	small match
 *			u32 index into wordoff_list; count given in cccccc
 *
 *	01000000	big match
 *			u32 index into wordoff_list; u32 count;
 *
 *	110ccccc	small exact match [can follow after normal match]
 *
 *	11000000	big exact match [dtto]
 *			u32 count;
 *
 *	001ccccc	near branching
 *			struct { u8 character; u16 rel_son_offset }  [ccccc];
 *
 *	00100000	far branching
 *			u8 count; struct { u8 character; u32 son_offset } [count];
 *
 *	11111111	vertex end [required]
 */

struct index_header {
	u32 magic;				/* INDEX_MAGIC_xxx */
	u32 version;				/* INDEX_VERSION */
	u32 num_phrases;			/* Number of phrases indexed */
	u32 num_wordoffs;			/* Number of individual words indexed */
	u32 num_fastfinds;			/* Number of fastfind trie items */
	u32 num_fastfind_vertices;		/* Number of trie vertices with precomputed fflist*/
	u32 timestamp;				/* Time when the index was built */
	u32 phrase_start;			/* Positions of parts of the index, byte offsets */
	u32 wordoff_start;
	u32 fastfind_start;
	u32 trie_start;
	u32 index_size;				/* Complete index bytesize */
} PACKED;

struct phrase {
	u32 relevance;				/* The phrase weight */
	struct phrase_attrs attrs;		/* defined in custom/custom.h */
	u16 text_len;
	byte text[1];				/* Phrase text in UTF-8 */
} PACKED;


#define INDEX_MAGIC 0x27021982		/* Finished index file */
#define INDEX_TEMP_MAGIC 0x06112005	/* Index under construction */

#define INDEX_FORMAT_VERSION 0x01	/* Current index format version */
#define INDEX_VERSION ((INDEX_FORMAT_VERSION << 16) | (CUSTOM_INDEX_TYPE << 8) | CUSTOM_INDEX_VERSION)

#endif
