/*
 * Kibitz Index Dumper
 *
 * (c) 2005-2012 Tomas Valla <tom@ucw.cz>
 *
 * This software may be freely distributed and used according to the terms
 * of the GNU Lesser General Public License. See COPYING for details.
 */

#include <kibitz/lib.h>
#include <ucw/conf.h>
#include <ucw/fastbuf.h>
#include <ucw/getopt.h>
#include <ucw/unaligned.h>
#include <kibitz/kibitz.h>
#include <kibitz/index.h>
#include <custom/custom.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>


static struct fastbuf *stdoutfb;
#define OUT(x...) bprintf(stdoutfb, x)

static struct fastbuf *idxfb;

static struct index_header idx_hdr;

static void section(char *name)
{
	static int section_cnt;
	if (section_cnt++)
		OUT("\n");
	OUT("%s:\n", name);
}

static void dump_header(void)
{
	if (idx_hdr.magic == INDEX_MAGIC)
		section("Index header");
	else
		section("Temporary index header");
	OUT("Index format version:\t\t0x%02x\n", idx_hdr.version >> 16);
	OUT("Custom index type:\t\t0x%02x version 0x%02x\n", (idx_hdr.version >> 8) & 0xff, idx_hdr.version & 0xff);
	time_t t = idx_hdr.timestamp;
	OUT("Build time:\t\t\t%s", ctime(&t));
	OUT("Number of phrases:\t\t%d\t(%d bytes)\n", idx_hdr.num_phrases, idx_hdr.wordoff_start-idx_hdr.phrase_start);
	OUT("Number of wordoffs:\t\t%d\t(%d bytes)\n", idx_hdr.num_wordoffs, idx_hdr.fastfind_start-idx_hdr.wordoff_start);
	OUT("Number of fastfinds:\t\t%d\t(%d bytes)\n", idx_hdr.num_fastfinds, idx_hdr.trie_start-idx_hdr.fastfind_start);
	OUT("Number of fastfind vertices:\t%d\n", idx_hdr.num_fastfind_vertices);

	OUT("Phrases start:\t\t\t0x%x\n", idx_hdr.phrase_start);
	OUT("Wordoffs start:\t\t\t0x%x\n", idx_hdr.wordoff_start);
	OUT("Fastfinds start:\t\t0x%x\n", idx_hdr.fastfind_start);
	OUT("Trie start:\t\t\t0x%x\n", idx_hdr.trie_start);
	OUT("Trie size:\t\t\t%d bytes\n", idx_hdr.index_size - idx_hdr.trie_start);
	OUT("Total index size:\t\t%d bytes\n", idx_hdr.index_size);
}

static void dump_phrases(void)
{
	section("Phrases");
	uns psize = idx_hdr.wordoff_start - idx_hdr.phrase_start;
	OUT("%d phrases total, %d bytes\n", idx_hdr.num_phrases, psize);
	bsetpos(idxfb, idx_hdr.phrase_start);
	byte *pbuf = xmalloc(psize);
	breadb(idxfb, pbuf, psize);

	struct phrase *pr;
	uns i, ptr = 0;

	for (i=0; i<idx_hdr.num_phrases; i++) {
		pr = (typeof(pr))(pbuf + ptr);
		OUT("phrase %d at 0x%x: '%s' text_len=%d relevance=%d custom: ", i, ptr,
			pr->text, pr->text_len, pr->relevance);
		dump_custom_struct_phrase_attrs(stdoutfb, &pr->attrs);
		OUT("\n");
		ptr += sizeof(struct phrase) + pr->text_len;
		while (ptr & 3)
			ptr++;
	}
	xfree(pbuf);
}

static void dump_wordoffs(void)
{
	section("Wordoffs");
	uns wsize = idx_hdr.fastfind_start - idx_hdr.wordoff_start;
	OUT("%d wordoffs total, %d bytes\n", idx_hdr.num_wordoffs, wsize);
	bsetpos(idxfb, idx_hdr.wordoff_start);
	u32 *wbuf = xmalloc(wsize);
	breadb(idxfb, wbuf, wsize);

	uns psize = idx_hdr.wordoff_start - idx_hdr.phrase_start;
	bsetpos(idxfb, idx_hdr.phrase_start);
	byte *pbuf = xmalloc(psize);
	breadb(idxfb, pbuf, psize);

	for (uns i=0; i<idx_hdr.num_wordoffs; i++) {
		uns wo = wbuf[i];
		byte *pt = ((struct phrase *) (pbuf+wo))->text;
		u32 pr = ((struct phrase *) (pbuf+wo))->relevance;
		OUT("wordoff %d: 0x%x (phrase '%s', relevance=%d)\n", i, wo, pt, pr);
	}
	xfree(wbuf);
	xfree(pbuf);
}

static void dump_fastfinds(void)
{
	section("Fastfinds");
	uns ffsize = idx_hdr.trie_start - idx_hdr.fastfind_start;
	OUT("%d fastfinds total, %d bytes\n", idx_hdr.num_fastfinds, ffsize);

	bsetpos(idxfb, idx_hdr.fastfind_start);
	u32 *ffbuf = xmalloc(ffsize);
	breadb(idxfb, ffbuf, ffsize);

	for (uns i=0; i<idx_hdr.num_fastfinds; i++)
		OUT("fastfind %d: phrase_off=0x%x quality=%d\n", i, ffbuf[2*i], ffbuf[2*i+1]);

	xfree(ffbuf);
}

static void dump_vertex(byte *trie, uns addr, uns level)
{
	byte *v = trie+addr;

	byte spc[2*level+1];
	memset(spc, ' ', 2*level);
	spc[2*level] = 0;

	byte br_cnt = 0;
	u32 br_buf[256];

	OUT("%sVERTEX at 0x%x:\n", spc, addr);
	for (;;) {
		uns opcode = *v++;

		if (opcode == 0xff) {
			OUT("%svertex_end\n", spc);
			break;
		}
		if ((opcode & 0xc0) == 0x80) {
			OUT("%sfastfind: index=%d items=%d\n", spc, GET_U32(v), opcode & 0x3f);
			v += 4;
			continue;
		}
		if ((opcode & 0xf0) == 0x10) {
			uns l = opcode & 15;
			byte buf[l + 1];
			memcpy(buf, v, l);
			buf[l] = 0;
			OUT("%sskip_char: %d bytes '%s'\n", spc, l, buf);
			v += l;
			continue;
		}
		if ((opcode & 0xc0) == 0x40) {
			if (opcode & 0x3f) {
				OUT("%ssmall_match: wref=%d cnt=%d\n", spc, GET_U32(v), opcode & 0x3f);
				v += 4;
			}
			else {
				OUT("%sbig_match: wref=%d cnt=%d\n", spc, GET_U32(v), GET_U32(v+4));
				v += 8;
			}
			continue;
		}
		if ((opcode & 0xc0) == 0xc0) {
			if (opcode & 0x1f)
				OUT("%ssmall exact match: cnt=%d\n", spc, opcode & 0x1f);
			else {
				OUT("%sbig exact match: cnt=%d\n", spc, GET_U32(v));
				v += 4;
			}
			continue;
		}
		if (opcode & 0x20) {
			if (opcode & 0x1f) {
				uns c = opcode & 0x1f;
				br_cnt = c;
				OUT("%snear_branching: cnt=%d\n", spc, c);
				for (uns i=0; i<c; i++) {
					uns ch = *v++;
					u16 ro = GET_U16(v);
					v += 2;
					br_buf[i] = addr+ro;
					OUT("%s branch %d: char=0x%x '%c' reloff=0x%x (absoff=0x%x)\n", spc,
						i, ch, ch, ro, addr+ro);
				}
			}
			else {
				uns c = *v++;
				br_cnt = c;
				OUT("%sfar_branching: cnt=%d\n", spc, c);
				for (uns i=0; i<c; i++) {
					uns ch = *v++;
					u32 ao = GET_U32(v);
					v += 4;
					br_buf[i] = ao;
					OUT("%s branch %d: char=0x%x '%c' absoff=0x%x\n", spc, i, ch, ch, ao);
				}
			}
			continue;
		}
		bflush(stdoutfb);
		ASSERT(0);
	}

	for (uns i=0; i<br_cnt; i++)
		dump_vertex(trie, br_buf[i], level+1);
}

static void dump_trie(void)
{
	section("Trie");
	uns ksize = idx_hdr.index_size - idx_hdr.trie_start;
	OUT("%d bytes total\n", ksize);
	byte *trie = xmalloc(ksize);
	bsetpos(idxfb, idx_hdr.trie_start);
	breadb(idxfb, trie, ksize);
	dump_vertex(trie, 0, 0);
	xfree(trie);
}

static void NONRET usage(byte *msg)
{
	if (msg)
		fputs(msg, stderr);
	fputs("\n\
Usage: kidxdump [<options>]\n\
-A		dump all\n\
-F		dump fastfinds\n\
-f <index>	index filename\n\
-h		dump header\n\
-T		dump trie\n\
-P		dump phrases\n\
-W		dump wordoffs\n\
", stderr);
	exit(1);
}
static char *options = "AFf:hPTW";

static int header_dump;
static int fastfinds_dump;
static int trie_dump;
static int phrases_dump;
static int wordoffs_dump;

int main(int argc, char **argv)
{
	int opt;

	stdoutfb = bfdopen(1, 65536);

	log_init(argv[0]);
	while ((opt = cf_getopt(argc, argv, options, CF_NO_LONG_OPTS, NULL)) >= 0)
		switch (opt) {
		case 'A':
			header_dump = 1;
			fastfinds_dump = 1;
			trie_dump = 1;
			phrases_dump = 1;
			wordoffs_dump = 1;
			break;
		case 'F':
			fastfinds_dump = 1;
			break;
		case 'f':
			idxfb = bopen(optarg, O_RDONLY, 65536);
			break;
		case 'h':
			header_dump = 1;
			break;
		case 'T':
			trie_dump = 1;
			break;
		case 'P':
			phrases_dump = 1;
			break;
		case 'W':
			wordoffs_dump = 1;
			break;
		default:
			usage("Bad options specified");
		}

	if (!idxfb)
		usage("No indexfile specified");

	/* Load header */
	breadb(idxfb, &idx_hdr, sizeof(idx_hdr));
	if (idx_hdr.magic != INDEX_MAGIC && idx_hdr.magic != INDEX_TEMP_MAGIC)
		die("Unrecognized index header (magic=%08x)", idx_hdr.magic);

	if (header_dump)
		dump_header();
	if (phrases_dump)
		dump_phrases();
	if (wordoffs_dump)
		dump_wordoffs();
	if (fastfinds_dump)
		dump_fastfinds();
	if (trie_dump)
		dump_trie();

	bclose(idxfb);
	bclose(stdoutfb);
	return 0;
}
